/***************************************************************************
*
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* copyright laws.
*
* ==========================================================================
*
* FILE: CUC_DString.cc
*
* AUTHOR:
*
* John Ehresman
* Edward Pattison-Gordon
* Stephan Deibel
*
* CREATION DATE: 
*
* 31 July 1995
*
* DESCRIPTION:
*
* A fairly simplistic dynamic string class.  Takes care of memory 
* management when setting string to a value or appending data to it.  Frees
* any heap allocated memory in the destructor.
*
* NOTES:
*
* Most member functions are trivial because this is based on CU_Text; and
* are therefore implemented within CU_DString.hh.
*
* MODIFICATIONS:
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
*
* $Log: CU_DString.cpp,v $
* Revision 1.2  1998/07/15 05:20:18  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
// Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
// Converting to CVS
//
* Revision 1.4  1996/03/18 04:43:42  jehresma
* Modified include's for dos filenames
*
* Revision 1.3  1996/03/13  16:11:20  jehresma
* win32 round 1
*
* Revision 1.2  1996/01/24  18:59:53  sdeibel
* Reimplemented using new CU_Text.*
*
*
*/


/**************************************************************************/
/* INCLUDES */

#ifdef dos
#include "CUString.hh"
#include "CUDebug.h"
#else
#include "CU_DString.hh"
#include "CU_Debug.h"
#endif

/****************************************************************************
* FUNCTION: CUC_DString::CUC_DString ()
*
* DESCRIPTION: Constructor for CUC_DString; sets all fields up for a blank
*  string.
*
* PARAMETERS: none
*
*/

CUC_DString::CUC_DString()
{
  CU_TXInit(&fText);
  CU_TXSetBlockSize(&fText, CUk_DefDStringBlockSize);
}


/****************************************************************************
* FUNCTION: CUC_DString::~CUC_DString ()
*
* DESCRIPTION: Destructor for CUC_DString; frees any heap allocated memory
*
* PARAMETERS: none
*
*/

CUC_DString::~CUC_DString()
{
  CU_TXClear(&fText);
}


/****************************************************************************
* FUNCTION: CUC_DString::TakeString ()
*
* DESCRIPTION: Returns a pointer to the C string owned by this DString
*  object.  The DString is cleared in the process and the caller takes
*  over memory ownership of the C string.
*
* PARAMETERS: none
*
* RETURNS: char *: The string; memory now owned by the caller
*
*/

char *
CUC_DString::TakeString ()
{
   char * temp_ptr;

   temp_ptr = (char *) CU_TXGetCStringPtr(&fText);
   CU_TXInit(&fText);
   return temp_ptr;
}


