/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
* 
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
* 
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
* 
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE:
* 
* CU_File.cpp
* 
* AUTHOR: 
* 
* Jaime Requena
* Stephan R.A. Deibel
* 
* CREATION DATE: 
* 
* 9.02.96
* 
* DESCRIPTION: 
* 
* File and file name manipulation utilities
*
* NOTES:
*
*
* MODIFICATIONS: 
* 
* $Log: CU_File.cpp,v $
* Revision 1.14  1998/12/05 00:04:55  sdeibel
* Added init function and fixed problems in CU_File unix impl
*
* Revision 1.13  1998/12/04 22:44:35  bsloan
* CreateWindow conflict, GetFileNamesInDir fix, PC gs_server now works
*
* Revision 1.12  1998/12/01 23:18:51  bsloan
* Port to MetroWerks CodeWarrior Pro4 for Windows
*
* Revision 1.11  1998/10/08 14:53:39  swang
* Changes to run in VC++, need to re-check compatibility with gnu linux
*
* Revision 1.10  1998/07/06 15:53:48  jehresma
* Fixed win32 impl of CU_GetFilenamesInDir
*
* Revision 1.9  1998/07/06 15:44:34  jehresma
* Defined CUt_StringSeq and support functions.  Defined CU_GetFilenamesInDir
* and CU_JoinPathname.
*
* Revision 1.8  1998/04/01 05:57:38  sdeibel
* Bug fix to CU_GetFileName()
*
* Revision 1.7  1998/03/17 22:01:35  sdeibel
* Misc params made const
*
* Revision 1.6  1998/03/17 21:37:12  jehresma
* Bug fix
*
* Revision 1.5  1998/03/17 20:59:01  jehresma
* Port to PC
*
* Revision 1.4  1998/03/17 19:27:14  sdeibel
* Relative/full path utils now work on UNIX
*
* Revision 1.3  1998/03/17 00:03:46  sdeibel
* Implemented relative path utils; not yet tested
*
* 
*/

#include "string.h"
#include "fcntl.h"

#ifdef unix
#include "dirent.h"
#endif /* unix */
#ifdef WIN32
#include "windows.h"
#endif /* WIN32 */

#include "CU_File.hh"
#include "CU_Debug.h"
#include "CU_Text.h"
#include <iostream.h> //SJW DEBUG


/*----------------------------------------------------------------------------*/

tError CU_CreateFile(const char* newname) {
  /* note: this will overwrite the file newname if it exists */
  FILE* the_file;
  the_file = fopen(newname, "w+");
  if (the_file != NULL) {
    fclose(the_file);
    return kNoError;
  }
  else 
    return kFileIOError;
}

/*----------------------------------------------------------------------------*/

tError CU_RemoveFile(const char* filename) {

  return remove(filename);

}

/*----------------------------------------------------------------------------*/

tError CU_CopyFile(const char* source, const char* dest) {

  FILE* tempfile;
  void* buffer;
  FILE* newfile;
  
  newfile = fopen(dest, "w+");
  if (newfile == NULL)
    return kFileIOError;
  tempfile = fopen(source, "r");
  if (tempfile != NULL) {
    buffer = CU_Malloc(32000);
    if (buffer == NULL)
      return kOutOfMemory;
    while (fread(buffer, 32000, 1, tempfile) != 0)
      fwrite(buffer, 32000, 1, newfile);
    fwrite(buffer, 32000, 1, newfile);
    CU_Free(buffer);
  }
  else
    return kFileIOError;
  if (fclose(newfile) != 0)
    return kFileIOError;
  if (fclose(tempfile) != 0)
    return kFileIOError;
  return kNoError;
}

/*----------------------------------------------------------------------------*/

tError CU_CopyFile(const char* source, FILE* dest) {

  FILE* tempfile;
  void* buffer;
  
  if (dest == NULL)
    return kFileIOError;
  tempfile = fopen(source, "r+b");
  if (tempfile != NULL) {
    buffer = CU_Malloc(32000);
    if (buffer == NULL)
      return kOutOfMemory;
    while (fread(buffer, 32000, 1, tempfile) != 0)
      fwrite(buffer, 32000, 1, dest);
    fwrite(buffer, 32000, 1, dest);
    CU_Free(buffer);
  }
  else 
    return kFileIOError;
  
  if (fclose(tempfile) != 0)
    return kFileIOError;
  return kNoError;
}

/*----------------------------------------------------------------------------*/

tError CU_CopyFile(FILE* source, const char* dest) {

  void* buffer;
  FILE* newfile;
  
  newfile = fopen(dest, "w+");
  if (newfile == NULL)
    return kFileIOError;
  if (source != NULL) {
    buffer = CU_Malloc(32000);
    if (buffer == NULL)
      return kOutOfMemory;
    while (fread(buffer, 32000, 1, source) != 0)
      fwrite(buffer, 32000, 1, newfile);
    fwrite(buffer, 32000, 1, newfile);
    CU_Free(buffer);
  }
  else
    return kFileIOError;
  if (fclose(newfile) != 0)
    return kFileIOError;
  return kNoError;
}

/*----------------------------------------------------------------------------*/

tError CU_ExchangeFiles(const char* file_1, const char* file_2) {

  char filename[100];

  strcpy(filename, CU_GetFilePath(file_1)); 
  strcat(filename, "#Tempfile");
  if (CU_CopyFile(file_1, filename) != kNoError)
    return kFileIOError;
  if (CU_RemoveFile(file_1) != kNoError)
    return kFileIOError;
  if (CU_CopyFile(file_2, file_1) != kNoError)
    return kFileIOError;
  if (CU_CopyFile(filename, file_2) != kNoError)
    return kFileIOError;
  return kNoError;
  
}

/*----------------------------------------------------------------------------*/

tError CU_ExchangeFiles(const char* file_1, FILE* file_2) {

  char filename[100];

  strcpy(filename, CU_GetFilePath(file_1)); 
  strcat(filename, "#Tempfile");
  if (CU_CopyFile(file_1, filename) != kNoError)
    return kFileIOError;
  if (CU_RemoveFile(file_1) != kNoError)
    return kFileIOError;
  if (CU_CopyFile(file_2, file_1) != kNoError)
    return kFileIOError;
  if (CU_CopyFile(filename, file_2) != kNoError)
    return kFileIOError;
  return kNoError;
  
}

/*----------------------------------------------------------------------------*/

tError CU_FileMove(const char* old_full_path, const char* new_file_path) {

#ifdef macintosh
  UNIMPLEMENTED;
#endif

#ifdef WIN32
  char* old_name;
  char filename[100];

  old_name = CU_GetFileName(old_full_path);
  strcpy(filename, new_file_path);
  strcat(filename, old_name);
  CU_Free(old_name);
  if (strcmp(filename, old_full_path) != 0) {
    if (rename(old_full_path, filename) == 0)
      return kNoError;
    else return kFileIOError;
  }
  return kNoError;
#endif

#ifdef unix
  FILE* tempfile;
  void* buffer;
  FILE* newfile;
  char* old_name;
  char filename[100];

  old_name = CU_GetFileName(old_full_path);
  strcpy(filename, new_file_path);
  strcat(filename, old_name);
  CU_Free(old_name);
  if (strcmp(filename, old_full_path) != 0) {
    newfile = fopen(filename, "w+");
    if (newfile == NULL)
      return kFileIOError;
    tempfile = fopen(old_full_path, "r");
    if (tempfile != NULL) {
      buffer = CU_Malloc(32000);
      if (buffer == NULL)
	return kOutOfMemory;
      while (fread(buffer, 32000, 1, tempfile) != 0)
	fwrite(buffer, 32000, 1, newfile);
      fwrite(buffer, 32000, 1, newfile);
      CU_Free(buffer);
    }
    else
      return kFileIOError;
    if (fclose(newfile) != 0)
      return kFileIOError;
    if (fclose(tempfile) != 0)
      return kFileIOError;
    
    return CU_RemoveFile(old_full_path);
  }
  return kNoError;
#endif    

}

/*----------------------------------------------------------------------------*/

tError CU_RenameFile(const char* old_full_path, const char* newname) {

#ifdef macintosh
  UNIMPLEMENTED;
#endif

#ifdef WIN32
  char* old_dir;
  char filename[100];
  old_dir = CU_GetFilePath(old_full_path);
  strcpy(filename, old_dir);
  CU_Free(old_dir);
  strcat(filename, newname);
  if (strcmp(filename, old_full_path) != 0) {
    if (rename(old_full_path, filename) == 0)
      return kNoError;
    else return kFileIOError;
  }
  return kNoError;
#endif

#ifdef unix
  FILE* tempfile;
  void* buffer;
  FILE* newfile;
  char* old_dir;
  char filename[100];

  old_dir = CU_GetFilePath(old_full_path);
  strcpy(filename, old_dir);
  CU_Free(old_dir);
  strcat(filename, newname);
  if (strcmp(filename, old_full_path) != 0) {
    newfile = fopen(filename, "w+");
    if (newfile == NULL)
      return kFileIOError;
    tempfile = fopen(old_full_path, "r");
    if (tempfile != NULL) {
      buffer = CU_Malloc(32000);
      if (buffer == NULL)
	return kOutOfMemory;
      while (fread(buffer, 32000, 1, tempfile) != 0)
	fwrite(buffer, 32000, 1, newfile);
      fwrite(buffer, 32000, 1, newfile);
      CU_Free(buffer);
    }
    else
      return kFileIOError;
    if (fclose(newfile) != 0)
      return kFileIOError;
    if (fclose(tempfile) != 0)
      return kFileIOError;
    
    return CU_RemoveFile(old_full_path);
  }
  return kNoError;
#endif    

}

/*----------------------------------------------------------------------------*/

tBoolean CU_FileExists(const char* filename) {

  FILE* tempfile;
  tempfile = fopen(filename, "r");
  if (tempfile == NULL)
    return false;
  else {
    fclose(tempfile);
    return true;
  }    
}

/*----------------------------------------------------------------------------*/

FILE* CU_TempFile() {

  return tmpfile();

}

/*----------------------------------------------------------------------------*/

FILE* CU_TempCopy (const char* source_file)
{
  tError error_code;
  FILE* tempfile;

  GOODMEM(source_file);

  tempfile = CU_TempFile();

  error_code = CU_CopyFile(source_file, tempfile);
  if (error_code == kNoError)
    return tempfile;
  else
    return NULL;
}


/*----------------------------------------------------------------------------*/

char* CU_GetFileName (const char* full_path) 
{
  char* the_name;
  tInt4 i;

  GOODMEM(full_path);

  /* Find begining of file name */
  i = strlen(full_path) - 1; 
  while (i >= 0 && full_path[i] != CUk_FilePathSeperator)
    i--;

  /* Special case : full path is just a file name */
  if (i < 0) {
    the_name = strdup(full_path);
    return the_name;
  }

  /* Copy file name only out of full path */
  the_name = strdup(full_path + i + 1);
  GOODMEM(the_name); /* Bad API design; ambiguous NULL retval [SRAD] */
  return the_name;
}


/*----------------------------------------------------------------------------*/

char* CU_GetFilePath (const char* full_path) {
  char* the_name;
  tInt4 i;

  GOODMEM(full_path);

  /* Find beginning of file name */
  i = strlen(full_path) - 1;
  while (i >= 0 && full_path[i] != CUk_FilePathSeperator)
    i--;
  if (i < 0)
    return NULL;

  /* Copy file name */
  the_name = (char *) CU_Malloc(i + 2);
  GOODMEM(the_name); /* Bad API design; ambiguous NULL retval [SRAD] */
  strncpy(the_name, full_path, i);
  the_name[i] = '\0';
  return the_name;
}


/*----------------------------------------------------------------------------*/
/* Returns a file name that starts with 'name_prefix'. */

char* CU_GetUniqueFileName (const char* name_prefix)
{
  tInt4 seconds;
  char* unique_file_name;
  char* the_time;

  GOODMEM(name_prefix);

  unique_file_name = (char*) CU_Malloc(100);
  if (unique_file_name == NULL)
    return NULL;

  /*generate a unique string*/

time((time_t *)&seconds);

//time(&seconds);


#ifdef HACK // Won't compile on Min gnu win32
  strcpy(unique_file_name, ltoa(seconds));
#endif // HACK UNIMPLEMENTED;

  /*make sure it won't be too long*/
  if ((strlen(unique_file_name) + strlen(name_prefix)) > 31) 
    CU_StrDeleteChar(unique_file_name, 0, 
		     (31 - (strlen(unique_file_name) + strlen(name_prefix))));
  the_time = strdup(unique_file_name);
  strcpy(unique_file_name, "/");
  strcat(unique_file_name, name_prefix);

  /*add the prefix*/
  strcat(unique_file_name, the_time);
  CU_Free(the_time);

  return unique_file_name;
}


/*------------------------------------------------------------------
 * Remove filename from a given full pathname and return that filename as
 * function value.  This also works to remove the last element of a 
 * pathname (i.e., if the file seperator is ":" on the native platform,
 * the "D:" in A:B:C:D: and the "C" in A:B:C).  The passed char * is 
 * left with the residual file seperator at end, so that result for 
 * 'A/B/C/D/' would be 'A/B/C/'
 *
 */

char *CU_RemoveFileName (char *full_path)
{
  tInt4 i;
  char *retval;

  GOODMEM(full_path);

  /* Locate the file name */
  i = strlen(full_path) - 1;
#ifdef MS_WIN
  while (i >= 0 && ((full_path[i] != CUk_FilePathSeperator &&
                     full_path[i] != '/') ||
		    i == (tInt4) strlen(full_path) - 1))
    i--;
#else
  while (i >= 0 && (full_path[i] != CUk_FilePathSeperator ||
		    i == (tInt4) strlen(full_path) - 1))
    i--;
#endif
  if (i < 0) {
    retval = (char *) CU_Malloc(1);
    retval[0] = '\0';
    return retval;
  }

  /* Copy file name and remove from input string */
  retval = (char *) CU_Malloc(strlen(full_path) - i);
  strcpy(retval, &full_path[i+1]);
  full_path[i+1] = '\0';

  /* Success */
  return retval;
}


/*------------------------------------------------------------------
* Remove volume name from a given full pathname and return that 
* volume name as function value.  This returns anything before
* the first file seperator field in the given string; on UNIX
* it returns the leading file path component, including the
* initial '/'.
*/

char *CU_RemoveVolName (char *remove_str)
{
  tInt4 i;
  char *retval;
  char tmp_str[512];

  GOODMEM(remove_str);

  /* Trivial case */
  if (strlen(remove_str) == 0) {
    remove_str[0] = '\0';
    return NULL;
  }

  /* Find extent of first step in given path */
  i = 0;
#ifdef unix
  while (i < (tInt4) strlen(remove_str) && 
	 (i == 0 || remove_str[i] != CUk_FilePathSeperator))
    i++;
#else
  while (i < (tInt4) strlen(remove_str) && 
	 (remove_str[i] != CUk_FilePathSeperator && remove_str[i] != '/'))
    i++;
#endif
  if (i >= (tInt4) strlen(remove_str)) {
    retval = (char *) CU_Malloc(1);
    retval[0] = '\0';
    return retval;
  }

  /* Build result */
  retval = (char *) CU_Malloc(i + 1);
  strncpy(retval, remove_str, i);
  retval[i] = '\0';
  strcpy(tmp_str, &remove_str[i]);
  strcpy(remove_str, tmp_str);

  /* Return value */
  return retval;
}


/*------------------------------------------------------------------
* Convert a full pathname to a partial relative pathname 
* (relative to the given file).
*
* The format for conversion is to as follows:
*
* convert_path and relative_to are sent as standard pathnames with
* CUk_FileSeperator as the path element seperator.  On MS Win, the volume
* name should be prepended as in the form C:\X\Y\Z).
*
* Result path is as follows in the following cases:
*
* rel    A:B:C:F1
*        (where A is the volume, B:C are folder paths, and F1 is the file)
* cvt    A:B:C:F2 -> .F2                          
*        ('.' indicates the relTo path of 'A:B:C')
*
* rel    A:B:C:F1
* cvt    A:B:F2 -> .^F2                         
*        ('.' indicates relTo path of 'A:B:C' and '^' indicates removal 
*         of 'C', 'F2' indicates concat)
*
* rel    A:B:F1
* cvt    A:B:C:F2 -> .C:F2                   
*        ('.' indicates 'A:B' relTo path, 'C:F2' indicates concat 
*         of that string)
*
* rel    A:B:C:F1
* cvt    A:B:D:E:F2 -> .^D:E:F2                
*        ('.' indicates 'A:B:C' relTo path, '^' strips 'C', rest is
*         concatenated)
*
* rel    A:B:C:F1
* cvt    X:Y:Z:F2 -> X:Y:Z:F2                
*        (full path name since there is no commonality)
*
* NOTE: '^' and '.' above are actually non-printable characters so file
* names may contain '.'s and '^'s
*
*/

void CU_FullToRelativePath (char *convert_path, const char *relative_to)
{
  char cvt_dotpath[512];
  char rel_dotpath[512];
  char *cvt_file;
  char *rel_file;
  char *cvt_step;
  char *rel_step;
  char common_path[512];
  char result[512];
  tUInt4 tmp_i;
    
  /* Check if paths are the same; if so indicate "self" in returned path */
  /* The algorithm below would instead include the filename in the */
  /* relative reference, so we bypass it */
  if (!strcmp(convert_path, relative_to)) {
    convert_path[0] = CUk_RelPathIndicator;
    convert_path[1] = '\0';
    return;
  }
  
  /* Set up file names and dotpaths */
  strcpy(cvt_dotpath, convert_path);
  strcpy(rel_dotpath, relative_to);
  cvt_file = CU_RemoveFileName(cvt_dotpath);
  rel_file = CU_RemoveFileName(rel_dotpath);

  /* Calculate the common path */
  cvt_step = CU_RemoveVolName(cvt_dotpath);
  rel_step = CU_RemoveVolName(rel_dotpath);
  common_path[0] = '\0';
  while (!strcmp(cvt_step, rel_step) && (strlen(cvt_step) > 0) &&
	 (strlen(rel_step) > 0)) {
    strcat(common_path, rel_step);
    if (cvt_step != NULL)
      CU_Free(cvt_step);
    if (rel_step != NULL)
      CU_Free(rel_step);
    cvt_step = CU_RemoveVolName(cvt_dotpath);
    rel_step = CU_RemoveVolName(rel_dotpath);
  }

  /* Build the relative path in the case that the relativeTo path is */
  /* shorter than the convert path (i.e., when the convert path indicates */
  /* a sub-directory of the relativeTo path) */
  if (strlen(rel_step) == 0) {

    /*Start with relative path indicator*/
    result[0] = CUk_RelPathIndicator;
    result[1] = '\0';

    /*Append remainder of convert path*/
    while (strlen(cvt_step) != 0) {
      strcat(result, cvt_step);
      if (cvt_step != NULL)
	CU_Free(cvt_step);
      cvt_step = CU_RemoveVolName(cvt_dotpath);
    }

    /*Append convert file name*/
    tmp_i = strlen(result);
    result[tmp_i] = CUk_FilePathSeperator;
    result[tmp_i+1] = '\0';
    strcat(result, cvt_file);
  }

  /* Build the relative path in the case that the convert path is */
  /* shorter than the relativeTo path (i.e., when the relativeTo */
  /* path indicates a sub-directory of the convert Path) */
  else if (strlen(cvt_step) == 0) {

    /* Start with relative path indicator */
    result[0] = CUk_RelPathIndicator;
    result[1] = '\0';

    /* Add any needed strip entry indicators */
    while (strlen(rel_step) > 0) {
      tmp_i = strlen(result);
      result[tmp_i] = CUk_StripEntryIndicator;
      result[tmp_i+1] = '\0';
      if (rel_step != NULL)
	CU_Free(rel_step);
      rel_step = CU_RemoveVolName(rel_dotpath);
    }

    /* Append convert file name */
    tmp_i = strlen(result);
    result[tmp_i] = CUk_FilePathSeperator;
    result[tmp_i+1] = '\0';
    strcat(result, cvt_file);
  }

  /* Build the relative path in the case that the convert path and */
  /* relativeTo path are entirely different - just use the full pathname */
  else if (strlen(common_path) == 0) {
    strcpy(result, convert_path);
  }

  /* Build the relative path in the case that the convert path and */
  /* relativeTo path share only a common set of ancestors) */
  else if (strcmp(cvt_step, rel_step) != 0) {

    /* Start with relative path indicator */
    result[0] = CUk_RelPathIndicator;
    result[1] = '\0';

    /* Add any needed strip entry indicators */
    while (strlen(rel_step) != 0) {
      tmp_i = strlen(result);
      result[tmp_i] = CUk_StripEntryIndicator;
      result[tmp_i+1] = '\0';
      if (rel_step != NULL)
	CU_Free(rel_step);
      rel_step = CU_RemoveVolName(rel_dotpath);
    }

    /* Add remainder of convert path */
    while (strlen(cvt_step) != 0) {
      strcat(result, cvt_step);
      if (cvt_step != NULL)
	CU_Free(cvt_step);
      cvt_step = CU_RemoveVolName(cvt_dotpath);
    }

    /* Add convert filename */
    tmp_i = strlen(result);
    result[tmp_i] = CUk_FilePathSeperator;
    result[tmp_i+1] = '\0';
    strcat(result, cvt_file);
  }

  /* Copy result to parameter */
  strcpy(convert_path, result);

  /* Clean up */
  if (cvt_step != NULL)
    CU_Free(cvt_step);
  if (rel_step != NULL)
    CU_Free(rel_step);
  if (cvt_file != NULL)
    CU_Free(cvt_file);
  if (rel_file != NULL)
    CU_Free(rel_file);
}


/*------------------------------------------------------------------*/
/* Convert a partial relative pathname generated by FullToRelativePath */
/* back into a standard full pathname */

void CU_RelativeToFullPath (char *convert_path, const char *relative_to)
{
  char dotpath[512];
  char *rel_file = NULL;
  char result[512];
  tInt4 i;
  char *tmp_str = NULL;

  /* Check if convert_path == CUk_RelPathIndicator.  If so, then this */
  /* relative path is just "self", so we return the value of relative_to */
  /* in convert_path */
  if ((strlen(convert_path) == 1) && (convert_path[0] == CUk_RelPathIndicator)) {
    strcpy(convert_path, relative_to);
    return;
  }

  /* Calculate the dot path */
  strcpy(dotpath, relative_to);
  rel_file = CU_RemoveFileName(dotpath);

  /* Add dot path to result if convertPath has a dot */
  if (convert_path[0] == CUk_RelPathIndicator)
    strcpy(result, dotpath);
  else
    result[0] = '\0';

  /* Remove steps off end of result for each strip entry indicator */
  /* found in convert path */
  i = 1;
  while ((convert_path[i] == CUk_StripEntryIndicator) && 
	 (i < (tInt4) strlen(convert_path))) {
    if (tmp_str != NULL)
      CU_Free(tmp_str);
    tmp_str = CU_RemoveFileName(result);
    i++;
  }

  /* If we found some commonality between paths - change the var */
  /* parameter convert_path.  If no common path was found then */
  /* convert_path is already the correct full path name */
  if (strlen(result) > 0) {
    strcat(result, &convert_path[i+1]);
    strcpy(convert_path, result);
  }
  
  /* Clean up */
  if (tmp_str != NULL)
    CU_Free(tmp_str);
  if (rel_file != NULL)
    CU_Free(rel_file);
}


/*----------------------------------------------------------------------------*/
/* Convert relative path (which contains embedded non-printable chars */
/* into a printable ascii form) */

char *CU_RelativeToAscii(const char *relative)
{
  tUInt4 i, j;
  char *retval = (char *) CU_Malloc(strlen(relative) * 2 + 1);

  // Check failure
  if (retval == NULL)
    return NULL;

  // Convert
  for (i = 0, j = 0; i < strlen(relative); i++) {
    if (relative[i] == CUk_RelPathIndicator) {
      retval[j] = '\\';
      retval[j+1] = '.';
      j = j + 2;
    }
    else if (relative[i] == CUk_StripEntryIndicator) {
      retval[j] = '\\';
      retval[j+1] = '^';
      j = j + 2;
    }
    else {
      retval[j] = relative[i];
      j++;
    }
  }

  // Success
  retval[j] = '\0';
  return retval;
}


/*----------------------------------------------------------------------------*/
/* Convert printable ascii form of a relative path into the canonical form, */
/* which contains embedded non-printable characters) */

char *CU_AsciiToRelative(const char *ascii)
{
  tUInt4 i, j;
  char *retval = (char *) CU_Malloc(strlen(ascii) + 1);

  // Check failure
  if (retval == NULL)
    return NULL;

  // Convert
  for (i = 0, j = 0; i < strlen(ascii); i++) {
    if (ascii[i] == '\\' && ascii[i+1] == '.') {
      retval[j] = CUk_RelPathIndicator;
      j++;
      i++;
    }
    else if (ascii[i] == '\\' && ascii[i+1] == '^') {
      retval[j] = CUk_StripEntryIndicator;
      j++;
      i++;
    }
    else {
      retval[j] = ascii[i];
      j++;
    }
  }

  // Success
  retval[j] = '\0';
  return retval;
}


/****************************************************************************
* FUNCTION:
*
* CU_GetFilenamesInDir()
*
* DESCRIPTION:
*
* List the names of all files and subdirectories in a directory.
* NOTE: Non-recursive--search is one directory deep.
*
* PARAMETERS:
*
* tError* err  -- Error code; must be kNoError on entry
* const char* dir_name -- Full path of directory to list
* 
* RETURNS:
*
* CUt_FileNameSeq -- Sequence of CUt_FileName, one for each entry found
*
*/

CUt_FileNameSeq CU_GetFilenamesInDir(tError* err, 
				     const char* dir_name)
{
  CUt_FileNameSeq name_seq;

  ASSERT(err != NULL && *err == kNoError);
  GOODMEM(dir_name);
  
  CU_SeqInit(&name_seq);
	CUt_FileName next_file; 

#ifdef unix
  {
    DIR* dir;
    struct dirent* entry;

    dir = opendir(dir_name);
    if ( dir == NULL ) {
      *err = kFileIOError;
      goto failure;
    }

    /* Read all entries and copy to name_seq */
    entry = readdir(dir);
    while (entry != NULL) {
      if (strcmp(entry->d_name, ".") && 
	  strcmp(entry->d_name, "..")) {
#ifdef BROKEN /* d_type doesn't exist! [SRAD]) */
	next_file.is_directory = (entry->d_type & DT_DIR)? TRUE : FALSE;
#else
	next_file.is_directory = FALSE;  // TEMPORARY [SRAD]
#endif
	next_file.name = CU_StrDup(entry->d_name);
	*err = CU_SeqAppend(&name_seq, &next_file, sizeof(CUt_FileName));
	if (*err != kNoError) {
	  closedir(dir);
	  goto failure;
	}
      }
      entry = readdir(dir);
    }

    closedir(dir);
  }

#elif defined(WIN32)

  {
    HANDLE find_handle;
    WIN32_FIND_DATA find_data;
    DWORD last_error;

    // Find the first entry in current directory (match anything):
    char* pattern = CU_JoinPathname(dir_name, "*");
    find_handle = FindFirstFile(pattern, &find_data);
    CU_Free(pattern);
    if (find_handle == INVALID_HANDLE_VALUE) {
      if (GetLastError() == ERROR_NO_MORE_FILES)
        return name_seq;
      else {
        *err = kFileIOError;
        goto failure;
      }
    }
    
    /* Find all entries and append to name_seq */
    do {
      // Skip nonsense entries:
      if (!strcmp(find_data.cFileName, ".") || 
	  !strcmp(find_data.cFileName, ".."))
	continue;
      
      // Add next file or subdirectory:
      if (find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
	next_file.is_directory = TRUE;
      else
	next_file.is_directory = FALSE;
      next_file.name = CU_StrDup(find_data.cFileName);
      *err = CU_SeqAppend(&name_seq, &next_file, sizeof(CUt_FileName));
      if (*err != kNoError) {
	FindClose(find_handle);
        goto failure;
      }    	
    } while (FindNextFile(find_handle, &find_data));
    
    /* Close search and see if a real error occured */
    last_error = GetLastError();
    FindClose(find_handle);
    if (last_error != ERROR_NO_MORE_FILES) {
      *err = kFileIOError;
      goto failure;
    }    
  }
#endif // !unix

  return name_seq;

failure:
  for (tUInt4 i = 0; i < name_seq._length; ++i) {
    CU_Free(name_seq._buffer[i].name);
  }
  CU_SeqClear(&name_seq);
  return name_seq;
}


/****************************************************************************
* FUNCTION:
*
* CU_JoinPathname()
*
* DESCRIPTION:
*
* Joins a path name with a filename if the filename is not an absolute name;
* Note that under win32, a filename beginning with a drive letter is
* considered an absolute name
*
* PARAMETERS:
*
* const char* dir_name  -- Path name
* const char* file_name -- File name
* 
* RETURNS:
*
* char* -- Newly allocation pathname or NULL iff memory error
*
* NOTES:
*
* This is modeled on the Python library function os.path.join.
*
*/
char*
CU_JoinPathname(const char* dir_name, const char* file_name)
{
  char* full_name;
  tUInt4 dir_len;
  tUInt4 file_len;

  /* Error check */
  GOODMEM(dir_name);
  GOODMEM(file_name);
  
  /* Check to see if file_name is absolute */
  if ( file_name[0] == CUk_FilePathSeperator )
    return CU_StrDup(file_name);

  dir_len = strlen(dir_name);
  file_len = strlen(file_name);

#ifdef WIN32
  /* Check for drive letter */
  if ( file_len >= 2 && file_name[1] == ':' )
    return CU_StrDup(file_name);
#endif /* WIN32 */

  full_name = (char*)CU_Malloc(dir_len + file_len + 2);
  if ( full_name == NULL )
    return NULL;
  strcpy(full_name, dir_name);

  // Append path seperator if needed
  if ( dir_len > 0 && file_len > 0 
       && full_name[dir_len - 1] != CUk_FilePathSeperator ) {
    full_name[dir_len] = CUk_FilePathSeperator;
    full_name[dir_len + 1] = '\0';
    dir_len++;
  }

  // Append file name
  strcpy(&full_name[dir_len], file_name);

  return full_name;
}

