/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE:
* 
* CUC_Mutex.cc
* 
* AUTHOR: 
* 
* Jaime Requena
* 
* CREATION DATE: 
* 
* March 1997
* 
* DESCRIPTION: 
* 
* Implemetation of Mutex.
* 
* NOTES: 
* 
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
*
* $Log: CU_Mutex.cpp,v $
* Revision 1.3  1998/07/15 05:20:19  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
* Revision 1.2  1998/06/30 18:26:09  sdeibel
* Misc fixes to support compilation under gcc 2.7.2
*
// Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
// Converting to CVS
//
* Revision 1.3  1997/04/07 15:15:12  jehresma
* Added include files
*
* Revision 1.2  1997/03/25 19:15:07  sdeibel
* Minor changes
*
*
*/


/**************************************************************************/
/* INCLUDES */

#include "CU_Basics.h"
#include "CU_Debug.h"
#include "CU_Mutex.hh"
#include "CU_Thread.h"


/**************************************************************************/
/* CLASSES */

/*------------------------------------------------------------------------*/
/* Member functions for class CUC_Mutex */
/*------------------------------------------------------------------------*/


/****************************************************************************
* FUNCTION:
*
* CUC_Mutex::CUC_Mutex()
*
* DESCRIPTION:
*
* Constructor for this class.
*
* PARAMETERS:
*
* CORBA_Environment *                   -- The environment.
*
* EXCEPTIONS:
*
* NONE.
*
* NOTES:
*
* This function creates/initializes an object instance to a blank state.
* 
* Further initialization should be performed with additional calls.
*
*/

CUC_Mutex::CUC_Mutex(CORBA_Environment *ev, CUC_Resource* resource)
{

  GOODMEM(ev);
  GOODMEM(resource);
  
  fRes = resource;
  resource->Mutex(ev, this);

}

/****************************************************************************
* FUNCTION:
*
* CUC_Mutex::~CUC_Mutex()
*
* DESCRIPTION:
*
* Destructor for this class.
*
* PARAMETERS:
*
* EXCEPTIONS:
*
* NONE.
*
* NOTES:
*
* This function deletes an object.
* 
*/

CUC_Mutex::~CUC_Mutex()
{

  fRes->Release(this);

}

/*------------------------------------------------------------------------*/
/* Member functions for class CUC_Resource */
/*------------------------------------------------------------------------*/

/****************************************************************************
* FUNCTION:
*
* CUC_Resource::CUC_Resource()
*
* DESCRIPTION:
*
* Constructor for this class.
*
* PARAMETERS:
*
* EXCEPTIONS:
*
* NONE.
*
* NOTES:
*
* This function creates/initializes an object instance to a blank state.
* 
* Further initialization should be performed with additional calls.
*
*/

CUC_Resource::CUC_Resource()
{
  fMutex = NULL;
}

/****************************************************************************
* FUNCTION:
*
* CUC_Resource::~CUC_Resource()
*
* DESCRIPTION:
*
* Destructor for this class.
*
* PARAMETERS:
*
* EXCEPTIONS:
*
* NONE.
*
* NOTES:
*
* This function deletes an object.
* 
*/

CUC_Resource::~CUC_Resource()
{
}

/****************************************************************************
* FUNCTION:
*
* CUC_Resource:Mutex()
*
* DESCRIPTION:
*
*
* PARAMETERS:
*
* EXCEPTIONS:
*
* NONE.
*
* NOTES:
*
*/

void CUC_Resource::Mutex(CORBA_Environment *ev, CUC_Mutex* mutex)
{
#ifndef CUk_NOTHREADS
  void* null_pointer = NULL;
  tUInt4 size;
#endif

  GOODMEM(ev);
  GOODMEM(mutex);

#ifdef CUk_NOTHREADS
  //BARF;
#else
  CUt_BlockOpt option;
  size = 4;
  option = CUk_UntilEqual;
  CU_BlockThread(CU_GetActiveThread(), CUk_UntilEqual, 
		 &fMutex, &null_pointer, &size, NULL, FALSE);
#endif
   
  fMutex = mutex;

}


/****************************************************************************
* FUNCTION:
*
* CUC_Resource:Release()
*
* DESCRIPTION:
*
* PARAMETERS:
*
* CUC_Mutex*              
*
* EXCEPTIONS:
*
* NONE.
*
* NOTES:
*
*/

void CUC_Resource::Release(CUC_Mutex* mutex)
{
  GOODMEM(mutex);

  ASSERT(fMutex == mutex);
  fMutex = NULL;

}
