/***************************************************************************
*
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_ThreadTest.c
* 
* AUTHOR: 
* 
* Stephan R.A. Deibel
*
* CREATION DATE: 
* 
* 8/2/94
*
* VERSION: 
* 
* 8/2/94
*
* DESCRIPTION: 
* 
* Test / sample program for the threads package.
*
* NOTES: 
*
* This is pretty brain-dead.  Normally, we would write this to use CG and
* be platform-independent (without all the #ifdef's) but in this case, we
* want to keep the CU code independent of higher level tools.
*
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* 8/2/94   Stephan   Created for testing.
* 
*/

#ifdef CUk_NOTHREADS
#error Cannot test threads when building with CUk_NOTHREADS!
#error You need to rebuild CU with this flag turned off!
#endif

#include <stdio.h>

#ifdef dos
#include "CUThread.h"
#include "CUBasics.h"
#include "CUDebug.h"
#include "CUTime.h"
#else 
#include "CU_Thread.h"
#include "CU_Basics.h"
#include "CU_Debug.h"
#include "CU_Time.h"
#endif /* dos */

#ifdef macintosh
#include <Events.h>
#include <AppleEvents.h>
#include <Dialogs.h>
#include <Gestalt.h>
#include <TextEdit.h>
#include <LowMem.h>
#ifdef GENERATINGCFM
#include <MixedMode.h>
#endif
#endif

#ifdef MS_WIN
#include <windows.h>
#endif

#ifdef macintosh
pascal OSErr DoAEOpenApplication(AppleEvent *, AppleEvent *, long);
pascal OSErr DoAEOpenDocuments(AppleEvent *, AppleEvent *, long);
pascal OSErr DoAEPrintDocuments(AppleEvent *, AppleEvent *, long);
pascal OSErr DoAEQuitApplication(AppleEvent *, AppleEvent *, long);
#endif

#define NUM_THREADS 6
#define RUN_TICKS 10 * 60

#define EVT_CHECK_FREQ 30

typedef struct {
  CUt_ThreadID thread_ID;
  tUInt4 flop_count;
  tUInt4 stop;
} tMasterData;

tMasterData gMasterData;

/* Routine to block master task except when an event is pending */
/* or it is time to quit */
tBoolean RemainBlocked(tMasterData *md, void *unused)
{
#ifdef macintosh
  EventRecord dummy;
  printf("Checking GUI block... Ticks = %ld; flop_count = %ld\n", 
         CU_CurrentTicks(), gMasterData.flop_count);
  if (CU_CurrentTicks() > gMasterData.stop)
    return(FALSE);
  else if (EventAvail(everyEvent, &dummy))
    return(FALSE);
  else
    return(TRUE);
#else
  printf("Checking GUI block... Ticks = %ld; flop_count = %ld\n", 
         CU_CurrentTicks(), gMasterData.flop_count);
  if (CU_CurrentTicks() > gMasterData.stop)
    return(FALSE);
  else
    return(TRUE);
#endif
}

/* Null thread ... just counts until it is nuked from the outside */
void EXPORT_PREFIX 
ThreadProc(void *data)
{
  tUInt4 *count = data;
   
  printf("Running thread procedure. \n" );

  for (;;) {
/*    printf("incrementing counter\n");*/
    (*count)++;
    CU_ThreadYield();
  }
}

#ifdef MS_WIN_GUI
int FAR PASCAL 
WinMain(HANDLE hInstance,       /* Application Instance Handle*/
        HANDLE hPrevInstance,   /* Previous Instance Handle*/
        LPSTR lpCmdLine,        /* Pointer to Command Line*/
        int nCmdShow)           /* Show Window Option*/
#else
#ifdef main
#undef main
#endif
int main()
#endif
{
  CUt_ThreadID threads[NUM_THREADS];
  CUt_ThreadID t_id;
  tUInt2 i;
  tUInt4 temp_long;
#ifdef macintosh
  EventRecord event;
  WindowPtr win;
  long temp_gestalt = 0;
  OSErr ret_code;

  /* Macintosh toolbox init */
  MaxApplZone();
  MoreMasters();
  MoreMasters();
#ifdef __MWERKS__
  InitGraf((Ptr) &qd.thePort);
#else
  InitGraf((Ptr) &thePort);
#endif
  InitFonts();
  InitWindows();
  InitMenus();
  TEInit();
  InitDialogs(0);
  FlushEvents(everyEvent, 0);
  InitCursor();

  /* Install AE handlers */
  if (Gestalt(gestaltAppleEventsAttr, &temp_gestalt) == noErr) {
    ret_code = AEInstallEventHandler(kCoreEventClass, kAEOpenApplication,
                    (AEEventHandlerUPP) NewAEEventHandlerProc(DoAEOpenApplication), 0, FALSE);
										
    if (ret_code == noErr)
      ret_code = AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments,
                    (AEEventHandlerUPP) NewAEEventHandlerProc(DoAEOpenDocuments), 0, FALSE);
	
    if (ret_code == noErr)
      ret_code = AEInstallEventHandler(kCoreEventClass, kAEPrintDocuments,
                    (AEEventHandlerUPP) NewAEEventHandlerProc(DoAEPrintDocuments), 0, FALSE);
	
    if (ret_code == noErr)
      ret_code = AEInstallEventHandler(kCoreEventClass, kAEQuitApplication,
                    (AEEventHandlerUPP) NewAEEventHandlerProc(DoAEQuitApplication), 0, FALSE);
		
    if (ret_code != noErr)
      ExitToShell();
  }
#endif /* macintosh */

  /* Init CU facility */
  if (CU_Initialize() != kNoError)
    UNIMPLEMENTED;

  /* Start up tasking */
  if ((gMasterData.thread_ID = CU_StartTasking(&gMasterData)) == 0)
    UNIMPLEMENTED;   

  /* Start up process yielder thread */
#ifdef macintosh
  CU_CreateThread(CU_ProcessYielder, NULL, (size_t)CUk_UseDefaultStackSize);
#endif

  /* Create a bunch of threads with default stack size */
  for (i = 0; i < NUM_THREADS; i++) {
/*  	 threads[i] = 0;*/
    t_id = CU_CreateThread(ThreadProc, &gMasterData.flop_count, 0);
/*
    if (threads[i] == 0) {
      UNIMPLEMENTED;
    }
*/
  }
  
  /* Run for a while */
  gMasterData.stop = CU_CurrentTicks() + RUN_TICKS;
  while (CU_CurrentTicks() < gMasterData.stop) {

#ifdef macintosh
	/* Get next event without yielding time to system (done elsewhere!) */
	if (GetNextEvent(everyEvent, &event)) {
		/* Process simple events */
		switch (event.what) {
			case mouseDown:
				if (FindWindow(event.where, &win) == inSysWindow)
					SystemClick(&event, win);
				break;
			case app4Evt:
				if ((event.message & 0xFF000000) == 0x01000000) {
					if (event.message & 0x00000001)
						CU_ApplStatusChanged(CUk_InForeground);
					else 
						CU_ApplStatusChanged(CUk_InBackground);
				}
				break;
			case kHighLevelEvent:
				if (AEProcessAppleEvent(&event))
					;						
				break;
			default:
				break;
		}
	}
#endif

    /* Block main thread until there is another event */
    temp_long = EVT_CHECK_FREQ;
    CU_BlockThread(gMasterData.thread_ID, CUk_UntilFctFALSE,
                   (void *) RemainBlocked, NULL,
                   &temp_long, NULL, FALSE);

    /* Let other threads run! */
		/*    printf("Master thread running; ticks = %d (end = %d)\n",
     CU_CurrentTicks(),	gMasterData.stop);*/
    CU_ThreadYield();
  }
  
  /* Stop tasking and kill threads */
  CU_StopTasking();
  
  /* NOTE: Number of iterations between all threads is now in */
  /* gMasterData.flop_count */
  printf("flop_count = %ld\n", gMasterData.flop_count);
  
  /* Done */
  exit(0);
}


#ifdef macintosh
pascal OSErr 
DoAEOpenApplication(AppleEvent* messagein, AppleEvent* reply, long refIn)
{
  return noErr;
}

pascal OSErr 
DoAEOpenDocuments(AppleEvent* messagein, AppleEvent* reply, long refIn)
{
  return errAEEventNotHandled;
}

pascal OSErr 
DoAEPrintDocuments(AppleEvent* messagein, AppleEvent* reply, long refIn)
{
  return errAEEventNotHandled;
}

pascal OSErr 
DoAEQuitApplication(AppleEvent* messagein, AppleEvent* reply, long refIn)
{
  /* Stop tasking and kill threads */
  CU_StopTasking();
  
  /* NOTE: Number of iterations between all threads is now in */
  /* gMasterData.flop_count */
  printf("flop_count = %ld\n", gMasterData.flop_count);
  
  /* Bail out from here. */
  exit(0);
  return noErr;
}
#endif


/* end of CU_ThreadTest.c */

