/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Time.c
* 
* AUTHOR: 
* 
* Stephan R.A. Deibel
*
* CREATION DATE: 
* 
* 12/16/93
*
* VERSION: 
* 
* 8/8/94
*
* DESCRIPTION: 
* 
* Basic cross-platform time-related utilities.
* 
* NOTES: 
* 
* Time value to abs time structure functions are incomplete -- they
* don't handle dates before 1 Jan 1601, the don't take into
* consideration the absence of a leap year every 100 years (except
* every 400 years), and they ignore any minute_offset value
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* 12/16/93 Stephan   Created file / initial implementation
* 8/8/94   Stephan   UNIX implementation
* 
* $Log: CU_Time.c,v $
* Revision 1.9  1998/12/05 00:04:56  sdeibel
* Added init function and fixed problems in CU_File unix impl
*
* Revision 1.8  1998/12/01 23:18:52  bsloan
* Port to MetroWerks CodeWarrior Pro4 for Windows
*
* Revision 1.7  1998/10/08 14:53:40  swang
* Changes to run in VC++, need to re-check compatibility with gnu linux
*
* Revision 1.6  1998/08/17 22:36:47  jehresma
* Fixed time to/from absolute time functions, except for limitations
* listed at top of file
*
* Revision 1.5  1998/07/15 05:20:23  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
* Revision 1.4  1998/06/18 20:25:29  jehresma
* Port to PC
*
* Revision 1.3  1998/06/17 20:13:41  jehresma
* More time functions
*
* Revision 1.2  1998/02/22 17:06:51  sdeibel
* Modern implementation with gettimeofday() on unix.  Note that this
* can also be used to easily implement the US timer functions that
* are currently unimplemented on unix.
*
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.9  1996/12/18 15:39:12  jrosenbe
* Ditched ^L & ^J at start of file
*
* Revision 1.7  1996/05/30  20:35:23  jehresma
* Rearranged position of EXPORT_PREFIX
*
* Revision 1.6  1996/03/18  04:46:17  jehresma
* Added windows.h include file under MS_WIN
*
* Revision 1.5  1996/02/20  19:15:44  sdeibel
* Fixed bug in CU_CurrentTicks()
*
* Revision 1.4  1996/01/23  23:32:43  sdeibel
* Misc changes associated with introducing CU_Text.*
*
 * Revision 1.3  1995/11/07  01:03:41  yjlee
 * Fixed GetCurrentTicks for MS_WIN.
 *
 * Revision 1.2  1995/07/31  16:11:41  yjlee
 * Integrated Threads code into MS Windows.
 *
 * Revision 1.1  1995/07/14  19:31:25  sdeibel
 * Initial revision
 *
* Revision 1.17  1995/07/06  16:40:51  yjlee
* Changed a BARF to UNFINISHED in gettime.
*
* Revision 1.16  1995/06/26  18:56:50  yjlee
* Removed compile warning messages.
*
Revision 1.15  1995/06/19  16:33:45  jehresma
Added #ifdef sunos4 around header file not found on HPUX

Revision 1.14  1995/06/07  23:18:32  sdeibel
Changed to __MWERKS__ preproc flag

Revision 1.13  1995/06/06  16:05:08  sdeibel
Changed [U]Int* to t[U]Int*

Revision 1.12  1995/06/02  18:04:40  maint
Reverted header names.

Revision 1.11  1995/06/02  14:22:26  maint
Changed header file names.

Revision 1.10  1995/05/25  22:07:49  yjlee
Added string.h

Revision 1.9  1995/04/25  19:07:14  yjlee
Fixed function misspellings.

*  */

/****************************************************************************
* CONFIG */

/* I think gettimeofday() is available on all supported UNIXes, but this */
/* controls insertion of old code that uses ftime() on UNIX if that is */
/* needed */
#ifdef unix
#define NEW_UNIX_TIME 1
#endif


/****************************************************************************
* INCLUDES */

#ifdef dos 
#include "CUBasics.h"
#include "CUDebug.h"
#include "CUTime.h"
#else
#include "CU_Basics.h"
#include "CU_Debug.h"
#include "CU_Time.h"
#endif /* dos */

#ifdef macintosh
#include <LowMem.h>
#include <string.h>
#endif /* macintosh */

#ifdef unix
#include <stdlib.h>
#ifdef NEW_UNIX_TIME
#include <sys/time.h>
#include <unistd.h>
#else
#include <sys/timeb.h>
#ifdef sunos4
#include <sys/stdtypes.h>
#endif /* sunos4 */
#endif
#endif /* unix */

#ifdef MS_WIN
#include <windows.h>
#include <time.h>
#endif


/****************************************************************************
* GLOBALS */

/* Locally used global for machines that don't have a tick time base */
#ifdef unix
#ifdef NEW_UNIX_TIME
static struct timeval CUg_StartTime = { 0, 0 };
#else
static struct timeb CUg_StartTime = { 0, 0, 0, 0 };
#endif
#endif /* unix */


/**************************************************************************/
/* CONSTANTS */

const tUInt2 CUk_DaysPerNormalYear = 365;
const tUInt2 CUk_DaysPerLeapYear = 366;
const tUInt2 CUk_DaysPerQuadYears = (3*365 + 366);
const tUInt8 CUk_TimeIncrPerSecond = 10000000;
/* CUk_TimeIncrPerMinute = 60*CUk_TimeIncrPerSecond */
const tUInt8 CUk_TimeIncrPerMinute = 60*10000000;
/* CUk_TimeIncrPerHour = 60*CUk_TimeIncrPerMinute */
const tUInt8 CUk_TimeIncrPerHour = (tUInt8)60*((tUInt8)60*(tUInt8)10000000);
/* CUk_TimeIncrPerDay = 24*CUk_TimeIncrPerHour */
#define CUk_TimeIncrPerDay ((tUInt8)24*(tUInt8)60*(tUInt8)60*(tUInt8)10000000)
/* CUk_TimeIncrPerRegYear = 365*CUk_TimeIncrPerDay */
const tUInt8 CUk_TimeIncrPerRegYear = 365*(24*((tUInt8)60*((tUInt8)60*(tUInt8)10000000)));
/* CUk_TimeIncrPerLeapYear = 366*CUk_TimeIncrPerDay */
const tUInt8 CUk_TimeIncrPerLeapYear = 366*(24*((tUInt8)60*((tUInt8)60*(tUInt8)10000000)));
const tUInt1 CUk_DaysPerMonth[12] = { // non-leap year; 0 = Jan
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

#define CUk_DayIndexOf1Jan1601 6652


/****************************************************************************
* FUNCTION:
*
* CU_InitTicks
*
* DESCRIPTION:
*
* Initialize the tick count facility.
*
* RETURNS:
*
* tBoolean    -- TRUE if succeeded; FALSE if not.
*
* NOTES:
*
* Only the first call will init local structures.  Subsequent
* calls are ignored and will return TRUE.
*
* ftime does not work properly on Linux and in fact is replaced by
* gettimeofday() on 
*/

tBoolean
CU_InitTicks()
{
#ifdef unix
#ifdef NEW_UNIX_TIME
  struct timezone tz;

  /* Init CUg_StartTime for UNIX only if first call */
  if ((CUg_StartTime.tv_sec == 0) && (CUg_StartTime.tv_usec == 0))
    if (gettimeofday(&CUg_StartTime, &tz) != 0)
      return(FALSE);
#else
  /* Init CUg_StartTime for UNIX only if first call */
  if ((CUg_StartTime.time == 0) && (CUg_StartTime.millitm == 0))
    if (ftime(&CUg_StartTime) == -1)
      return(FALSE);
#endif
#endif /* unix */

  /* If initted already or succeeded, return TRUE */
  return( TRUE );

} /* InitTicks */

 
/****************************************************************************
* FUNCTION:
*
* CU_CurrentTicks
*
* DESCRIPTION:
*
* Get the current tick count (ticks = 1/60 sec).
*
* RETURNS:
*
* tUInt4     -- Current tick count, or 0 if an error occured.
*
* NOTES:
*
* You *must* call CU_Initialize() (which calls CU_InitTicks()) before 
* calling this function
*
* On most platforms, this will be the number of ticks (1/60 sec) since 
* reboot or process startup.  
*
* Ticks are intended for use in relative timing only.
*
*/

tUInt4 EXPORT_PREFIX
CU_CurrentTicks()
{
#ifdef MS_WIN
  tUInt4 ret_val;
#endif
#ifdef unix
#if defined (linux)  || defined (hpux)
  struct timeval tv;
  struct timezone tz;
#else
  struct timeb curr_time;
#endif
  tUInt4 retval;
#endif /* unix */

  /* Debug mode error check */
#ifdef DEBUG
#ifdef unix
#ifdef NEW_UNIX_TIME
  if ((CUg_StartTime.tv_sec == 0) && (CUg_StartTime.tv_usec == 0)) {
    DEBUGMSG("Time not yet initialized");
  }
#else
  if ((CUg_StartTime.time == 0) && (CUg_StartTime.millitm == 0)) {
    DEBUGMSG("Time not yet initialized");
  }
#endif
#endif /* unix */
#endif /* DEBUG */

#ifdef macintosh
  /* Note that LMGetTicks() is much faster than TickCount() trap! */
  return LMGetTicks();
#endif

#ifdef unix
#ifdef NEW_UNIX_TIME
  if (gettimeofday(&tv, &tz) != 0)
    return(0);
  else {
    retval = 60 * (tv.tv_sec - CUg_StartTime.tv_sec);
    if (tv.tv_usec > CUg_StartTime.tv_usec)
      retval = retval + 
        (((tv.tv_usec - CUg_StartTime.tv_usec) * 60.0) / 1000000.0);
    else
      retval = retval +
        (((CUg_StartTime.tv_usec - tv.tv_usec) * 60.0) / 1000000.0);
    return(retval);
  }
#else
  /* Get current time and return difference in ticks */
  if (ftime(&curr_time) == -1)
    return(0);
  else {
    retval = 60 * (curr_time.time - CUg_StartTime.time);
    if (curr_time.millitm > CUg_StartTime.millitm)
      retval = retval + 
        (((curr_time.millitm - CUg_StartTime.millitm) * 60.0) / 1000.0);
    else
      retval = retval +
        (((CUg_StartTime.millitm - curr_time.millitm) * 60.0) / 1000.0);
    return(retval);
  }
#endif
#endif

#ifdef MS_WIN    
  ret_val = GetTickCount() * (tUInt4) 3 / (tUInt4) 50;
  return( ret_val );
#endif /* MS_WIN */

} /* CurrentTicks */


/****************************************************************************
* FUNCTION:
*
* CU_CreateUSTimer()
*
* DESCRIPTION:
*
* Create a new microsecond timer
*
* RETURNS:
*
* CUt_USTimer *      -- The newly created timer (an opaque platform-dependent
*                       structure).
*
* NOTES:
*
* Timer will not start running until you call CU_StartUSTimer().
*
*/

CUt_USTimer *
CU_CreateUSTimer()
{
  static tUInt4 timer_ID_seed = 0;
  CUt_USTimer *this_timer;
  
  UNTESTED;
  
  /* 
  * Create and initialize timer struct 
  */
  
  this_timer = (CUt_USTimer *) CU_Malloc(sizeof(CUt_USTimer));
  if (this_timer == NULL)
    return NULL;
  this_timer->timer_ID = ++timer_ID_seed;
  this_timer->running = FALSE;
  
  /* 
  * Platform-dependent initialization 
  */
  
#ifdef macintosh
  /* Calibrate by timing time manager overhead */
  memset((void *) &this_timer->tm_task, 0, sizeof(TMTask));
  InsTime((QElemPtr) &this_timer->tm_task);
  PrimeTime((QElemPtr) &this_timer->tm_task, -CUk_MaxUSDelay);
  RmvTime((QElemPtr) &this_timer->tm_task);
  this_timer->overhead = CUk_MaxUSDelay + this_timer->tm_task.tmCount;
#endif

#ifdef unix
  UNIMPLEMENTED;
#endif

#ifdef MS_WIN
  UNIMPLEMENTED;
  
/* For windows, a timer can be based on assembly language code that
   accesses Window's 840 ns resolution timer (this is based on the
   standard PC timer 0, which is used to generate the default 55 ms
   clock interrupt on all PCs (this is every 65536 flops of the 840 ns
   clock; that's 18.21 clock interrupts per second).  Anyway, Windows
   apparently reprograms that timer so we can get at this 840 ns
   resolution time base.
   
   To do that, you use the following two assembly language routines:
   
   getapi_
     proc near
     mov ax,1684h    ; Subcode to return API
     mov bx,05h      ; VTD ID
     xor di, di      ; Clear ES:DI
     mov es,di
     int 2fh
     mov dx,es
     mov ax,di
     ret
   getapi_ endp
   
   This routine fetches the entry point of the VTD.  It returns the
   address in DX:AX.  It calls INT 2FH, which is a kind of clearing
   house interrupt that returns the API entry point for all the device
   drivers Windows knows of.  You just plug a virtual device ID into
   the BX register (the ID of the VTD is 05) and call the interrupt;
   ES:DI returns holding the entry-point address in segment:offset
   form (or all zeros if the VxD has no entry point).
   
   gettick_
     proc near
     push cs
     mov bx,offset retspot
     push bx
     push dx        ; Push VTD segment
     push ax        ; Push VTD offset
     mov ax, 100h   ; Function code
     retf           ; Make the far call
   retstop:
     mov dword ptr _tarray,eax
     mov dword ptr _tarray+4,edx
     ret
   gettick_ endp
   
   This routine calls the VTD.  The entry point is passed to the
   routine in the DX:AX register pair. Note that the routine fakes
   a CALL using a far return.  Also, this routine presumes tarray
   is a two-element array of double words.  The value returned is
   the number of 840ns ticks because Windows was started in 
   EDX:EAX.
   
   Both of these routines came from the April '95 BYTE magazine,
   pp. 215-216.  I just transcribed them from there and have no
   idea what I am talking about; but hope that this info will
   help some poor slob implement the US timing facility on the
   PC as well!
*/
#endif

  /*
  * Return new timer
  */
  
  return this_timer;
}


/****************************************************************************
* FUNCTION:
*
* CU_StartUSTimer()
*
* DESCRIPTION:
*
* Start a microsecond timer
*
* PARAMETERS:
*
* CUt_USTimer *      -- The timer to start
*
* NOTES:
* 
* The timer must not already be running when this routine is called!
* This is a program error!
*
*/

void EXPORT_PREFIX
CU_StartUSTimer(CUt_USTimer *timer)
{
  ASSERT(timer != NULL);
  ASSERT(timer->running == FALSE);
  
  UNTESTED;
  
  /*
  * Mark timer as running
  */
  
  timer->running = TRUE;
  
  /* 
  * Platform-dependent setup 
  */
  
#ifdef macintosh
  InsTime((QElemPtr) &timer->tm_task);
  PrimeTime((QElemPtr) &timer->tm_task, -CUk_MaxUSDelay);
#endif

#ifdef unix
  UNIMPLEMENTED;
#endif

#ifdef MS_WIN
  UNIMPLEMENTED;
#endif
}


/****************************************************************************
* FUNCTION:
*
* CU_CheckUSTimer()
*
* DESCRIPTION:
*
* Check the given US timer and return the time elapsed in micro seconds
* since it was started.
*
* PARAMETERS:
*
* CUt_USTimer *      -- The US timer
*
* RETURNS:
*
* tUInt4             -- The microseconds elapsed, or 0 if the timer has expired
*
* NOTES:
* 
* The timer must be running when this routine is called; checking a stopped timer
* is a program error!
*
* Actual timer resolution is platform-dependent:
*
* Macintosh -- 20 microseconds.
* Unix -- UNIMPLEMENTED
* MS Windows -- UNIMPLEMENTED
*
*/

tUInt4 EXPORT_PREFIX 
CU_CheckUSTimer(CUt_USTimer *timer)
{
  ASSERT(timer != NULL);
  ASSERT(timer->running == TRUE);
  
  UNTESTED;
  
#ifdef macintosh
  if (timer->tm_task.tmCount == 0)
    return 0;
  else
    return(CUk_MaxUSDelay + timer->tm_task.tmCount - timer->overhead);
#endif

#ifdef unix
  UNIMPLEMENTED;
  return 0;
#endif

#ifdef MS_WIN
  UNIMPLEMENTED;
  return 0;
#endif
}


/****************************************************************************
* FUNCTION:
*
* CU_StopUSTimer()
*
* DESCRIPTION:
*
* Stop the given US timer and return the time elapsed in micro seconds
* since it was started.
*
* PARAMETERS:
*
* CUt_USTimer *      -- The US timer
*
* RETURNS:
*
* tUInt4             -- The microseconds elapsed, or 0 if the timer had expired
*
* NOTES:
* 
* The timer must be running when this routine is called; stopping a timer
* more than once in a row is a program error!
*
* Actual timer resolution is platform-dependent:
*
* Macintosh -- 20 microseconds.
* Unix -- UNIMPLEMENTED
* MS Windows -- UNIMPLEMENTED
*
*/

tUInt4 EXPORT_PREFIX
CU_StopUSTimer(CUt_USTimer *timer)
{
  ASSERT(timer != NULL);
  ASSERT(timer->running == TRUE);
  
  UNTESTED;
  
  /*
  * Stop the timer
  */
  
  timer->running = FALSE;
  
  /*
  * Platform-dependent stop & computation code
  */
  
#ifdef macintosh
  RmvTime((QElemPtr) &timer->tm_task);
  if (timer->tm_task.tmCount == 0)
    return 0;
  else
    return(CUk_MaxUSDelay + timer->tm_task.tmCount - timer->overhead);
#endif

#ifdef unix
  UNIMPLEMENTED;
  return 0;
#endif

#ifdef MS_WIN
  UNIMPLEMENTED;
  return 0;
#endif
}


/****************************************************************************
* FUNCTION:
*
* CU_DestroyUSTimer()
*
* DESCRIPTION:
*
* Destroy the given US timer, freeing memory associated with it.
*
* PARAMETERS:
*
* tUInt4      -- The US timer ID, as returned previously from CreateUSTimer()
*
* NOTES:
*
* The timer must be stopped before it is destroyed; destroying a running
* timer is a program error!
*
*/

void EXPORT_PREFIX
CU_DestroyUSTimer(CUt_USTimer *timer)
{
  ASSERT(timer != NULL);
  ASSERT(timer->running == FALSE);
  
  UNTESTED;
  
  /*
  * Platform-dependent destroy code
  */
  
#ifdef unix
  UNIMPLEMENTED;
#endif

#ifdef MS_WIN
  UNIMPLEMENTED;
#endif

  /*
  * Destroy master structure 
  */
  
  CU_Free(timer);
}


/****************************************************************************
* FUNCTION:
*
* CU_CurrentUtcTime()
*
* DESCRIPTION:
*
* Retrieves current UTC time as a CUt_TimeValue
*
* PARAMETERS:
*
* none
* 
* RETURNS:
*
* CUt_TimeValue -- current time (or the time that the OS reported a few
*                  instruction cycles ago)
*
*/
CUt_TimeValue CU_CurrentUtcTime()
{
#ifdef WIN32
  {
    SYSTEMTIME sys_time;
    tUInt8 file_time;
    /* MS uses the same 100 nanosecond increment system, but uses
       1-1-1601 as it's base date and we use 10-15-1582 */
    static const tUInt8 incr_diff = CUk_DayIndexOf1Jan1601 * CUk_TimeIncrPerDay;

    ASSERT(sizeof(tUInt8) == sizeof(FILETIME));
    GetSystemTime(&sys_time);
    SystemTimeToFileTime(&sys_time, (FILETIME*)&file_time);
    return file_time + incr_diff;
  }
#endif /* WIN32 */

#ifdef unix
  {
    struct timeval tv;
    struct timezone tz;
    /* gettimeofday uses 1-1-1970 as its base, so we pre-calculate the
       time value of 1-1-1970 here. 1-1-1970 is 387 years (inluding 97
       leap years) + 77 days or 141332 days after 10-15-1582. */
    static const tUInt8 k1970 = (tUInt8)141795 * CUk_TimeIncrPerDay;
    CUt_TimeValue now = k1970;

    if (gettimeofday(&tv, &tz) != 0)
      return 0;

    /* Adjust for timezone and DST */
    tv.tv_sec += (timezone * 60 + daylight*60*60);

    now += (tv.tv_sec * CUk_TimeIncrPerSecond);
    now += (tv.tv_usec * 10);
    return now;
  }
#endif /* unix */
}


#ifdef MS_WIN
static tUInt1 CU_DaysPerMonth(tUInt2 month, tBoolean is_leapyear)
#else
static inline tUInt1 CU_DaysPerMonth(tUInt2 month, tBoolean is_leapyear)
#endif
{
  if ( is_leapyear && month == 1 )
    return CUk_DaysPerMonth[1] + 1;
  else
    return CUk_DaysPerMonth[month];
}

/****************************************************************************
* FUNCTION:
*
*
* DESCRIPTION:
*
*
* PARAMETERS:
*
* 
* RETURNS:
*
*
*/
CUt_TimeValue CU_AbsDateToTime(const CUt_AbsDate* abs_date)
{
  tUInt8 days; /* # of days to encode in return value */
  tUInt1 i;
  CUt_TimeValue retval;
  tBoolean is_leapyear = ((abs_date->year % 4) == 0);
  tUInt4 year = abs_date->year;

  GOODMEM(abs_date);

  if ( year <= 1600 )
    UNIMPLEMENTED;
  
  /* Calculate # of days between year 1600 and date in abs_date */
  year -= 1600;
  days = CUk_DayIndexOf1Jan1601 + year * CUk_DaysPerNormalYear;
  /* Add in leap years before this one */
  if ( year > 0 )
    days += ((year - 1) / 4 + 1);
  /* Add whole months before the current one */
  for ( i = 0; i < abs_date->month - 1; i++ )
    days += CU_DaysPerMonth(i, is_leapyear);
  if ( abs_date->day != 0 )
    days += abs_date->day - 1;

  /* Calculate return value */
  retval = (days * CUk_TimeIncrPerDay 
            + abs_date->hour * CUk_TimeIncrPerHour
            + abs_date->minute * CUk_TimeIncrPerMinute
            + abs_date->second * CUk_TimeIncrPerSecond
            + abs_date->nanosecond / 100);

  /* Adjust for minute_offset */
  retval += abs_date->minute_offset * CUk_TimeIncrPerMinute;

  return retval;
}


/****************************************************************************
* FUNCTION:
*
* CU_TimeToAbsDate()
*
* DESCRIPTION:
*
* Transforms a time structure to AbsDate
* (SJW: Is this function supposed to return local time or UTC?)
*
* PARAMETERS:
*
* 
* RETURNS:
*
*
*/

CUt_AbsDate CU_TimeToAbsDate(CUt_TimeValue time, tInt2 minute_offset)
{
  tUInt8 days; /* # of days */
  tUInt1 i;
  CUt_AbsDate retval;
  tUInt8 quad_years;
//  tUInt8 years;
  tBoolean is_leapyear;

  /* Extract number of days */
  days = time / CUk_TimeIncrPerDay;
  time -= days * CUk_TimeIncrPerDay;

  // Subtract 1601-1-1 - 1582-10-15 (6570 + 5 + 77 = 6652) to obtain
  // days since 1601-1-1
  if ( days < CUk_DayIndexOf1Jan1601 ) {
    UNIMPLEMENTED;
  }
  else {
    days -= CUk_DayIndexOf1Jan1601;

    /* Fill in year; start with 4 year blocks and then count up to 
       3 remaining */
    quad_years = days / CUk_DaysPerQuadYears;
    days -= quad_years * CUk_DaysPerQuadYears;
    retval.year = 1600 + (tUInt2) quad_years * 4;
    if ( days > CUk_DaysPerLeapYear ) {
      retval.year++;
      days -= CUk_DaysPerLeapYear;
    }
    for ( i = 0; days >= CUk_DaysPerNormalYear; i++ ) {
      ASSERT(i < 3);
      retval.year++;
      days -= CUk_DaysPerNormalYear;
    }

    /* Count months */
    is_leapyear = ((retval.year % 4) == 0);
    retval.month = 1;
    for ( i = 0; days >= CU_DaysPerMonth(i, is_leapyear); i++ ) {
      ASSERT(i < 11);
      retval.month++;
      days -= CU_DaysPerMonth(i, is_leapyear);
    }
    ASSERT(days < CU_DaysPerMonth(i, is_leapyear));
    retval.day = (tUInt1) days + 1;
  }

  /* Add minute adjustment */
  // SJW thinks next line shouldn't be there!
//  minute_offset += minute_offset * CUk_TimeIncrPerMinute;

  // And shouldn't we be adding this into retval, i.e., if we're returning local time?
  retval.minute_offset = minute_offset;

  /* Extract hours, minutes, seconds, & nanoseconds */
  // SJW added the typecasting
  retval.hour = (tUInt1) (time / CUk_TimeIncrPerHour);
  time -= retval.hour * CUk_TimeIncrPerHour;
  retval.minute = (tUInt1) (time / CUk_TimeIncrPerMinute);
  time -= retval.minute * CUk_TimeIncrPerMinute;
  retval.second = (tUInt1) (time / CUk_TimeIncrPerSecond);
  time -= retval.second * CUk_TimeIncrPerSecond;
  ASSERT(time < CUk_TimeIncrPerSecond);
  retval.nanosecond = (tUInt4) (time * 100);

  return retval;
}


/****************************************************************************
* FUNCTION:
*
* CU_GetCurrentMinuteOffset()
*
* DESCRIPTION:
*
* Gets the current minute offset of the local timezone -- adjusts for DST
*
* PARAMETERS: none
* 
* RETURNS:
*
* tInt2
*
*/
tInt2 CU_GetCurrentMinuteOffset()
{
#ifdef unix
  tzset();
  return timezone / 60 + (daylight * 60);
#endif /* unix */

#ifdef WIN32 
	#ifdef __MWERKS__
		time_t now = time(0);
		struct tm* utc = gmtime(&now);
		struct tm* local = localtime(&now);
		double delta = difftime(mktime(utc), mktime(local));
		return (int) (delta / 60);
	#else
	  _tzset();
	  return _timezone / 60 + (_daylight * 60);
	#endif /* __MWERKS__ */
#endif /* WIN32 */


}


/****************************************************************************
* FUNCTION:
*
* CU_AddYearsToTimeValue()
*
* DESCRIPTION:
*
* Adds given number of years to time value; will check for 
* underflow conditions so time value will be 0 
*
* PARAMETERS:
*
* CUt_TimeValue* tv
* tInt4 years
* 
* RETURNS:
*
* void
*
*/
void
CU_AddYearsToTimeValue(CUt_TimeValue* tv, tInt4 years)
{
  /* Error check */
  GOODMEM(tv);
  
  if ( years < 0 ) {
    CUt_TimeValue tmp = (-years) * CUk_TimeIncrPerRegYear;
    if ( tmp > *tv )
      *tv = 0;
    else
      *tv -= tmp;
  }
  else {
    *tv += years * CUk_TimeIncrPerRegYear;
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_AddDaysToTimeValue()
*
* DESCRIPTION:
*
* Adds given number of years to time value; will check for 
* underflow conditions so time value will be 0 
*
* PARAMETERS:
*
* CUt_TimeValue* tv
* tInt4 days
* 
* RETURNS:
*
* void
*
*/
void
CU_AddDaysToTimeValue(CUt_TimeValue* tv, tInt4 days)
{
  /* Error check */
  GOODMEM(tv);
  
  if ( days < 0 ) {
    CUt_TimeValue tmp = (-days) * CUk_TimeIncrPerDay;
    if ( tmp > *tv )
      *tv = 0;
    else
      *tv -= tmp;
  }
  else {
    *tv += days * CUk_TimeIncrPerDay;
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_AddSecondsToTimeValue()
*
* DESCRIPTION:
*
* Adds given number of seconds to time value; will check for 
* underflow conditions so time value will be 0 
*
* PARAMETERS:
*
* CUt_TimeValue* tv
* tInt4 seconds
* 
* RETURNS:
*
* void
*
*/
void
CU_AddSecondsToTimeValue(CUt_TimeValue* tv, tInt4 seconds)
{
  /* Error check */
  GOODMEM(tv);
  
  if ( seconds < 0 ) {
    CUt_TimeValue tmp = (-seconds) * CUk_TimeIncrPerSecond;
    if ( tmp > *tv )
      *tv = 0;
    else
      *tv -= tmp;
  }
  else {
    *tv += seconds * CUk_TimeIncrPerSecond;
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_AddNanoSecondsToTimeValue()
*
* DESCRIPTION:
*
* Adds given number of nanoseconds to time value; will check for 
* underflow conditions so time value will be 0 
*
* PARAMETERS:
*
* CUt_TimeValue* tv
* tInt8 seconds
* 
* RETURNS:
*
* void
*
*/
void
CU_AddNanoSecondsToTimeValue(CUt_TimeValue* tv, tInt8 nanoseconds)
{
  /* Error check */
  GOODMEM(tv);
  
  if ( nanoseconds < 0 ) {
    CUt_TimeValue tmp = (-nanoseconds/100);
    if ( tmp > *tv )
      *tv = 0;
    else
      *tv -= tmp;
  }
  else {
    *tv += nanoseconds/100;
  }
}


tUInt4 CU_GetYearsOfTimeValue(CUt_TimeValue tv)
{
//SJW added typecasting
  tUInt4 retval = (tUInt4) (tv / CUk_TimeIncrPerRegYear);
  ASSERT((tv / CUk_TimeIncrPerRegYear) < kMaxUInt4);
  return retval;
}

tUInt4 CU_GetDaysOfTimeValue(CUt_TimeValue tv)
{
//SJW added typecasting
  tUInt4 retval = (tUInt4) (tv / CUk_TimeIncrPerDay);
  ASSERT((tv / CUk_TimeIncrPerDay) < kMaxUInt4);
  return retval;
}

tUInt4 CU_GetSecondsOfTimeValue(CUt_TimeValue tv)
{
//SJW added typecasting
  tUInt4 retval = (tUInt4) (tv / CUk_TimeIncrPerSecond);
  ASSERT((tv / CUk_TimeIncrPerSecond) < kMaxUInt4);
  return retval;
  
}

tUInt8 CU_GetNanoSecondsOfTimeValue(CUt_TimeValue tv)
{
  return tv * 100;
}

/* end of CU_Time.c */
